/*
 * Copyright 1995-1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <MapUnidraw/mapclasses.h>
#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapexport.h>
#include <MapUnidraw/mapps.h>
#include <MapUnidraw/maproute.h>
#include <MapUnidraw/mapscripts.h>
#include <MapUnidraw/mapviewer.h>
#include <MapUnidraw/mapviews.h>

#include <Map/mapdb.h>
#include <Map/maplib.h>
#include <Map/mapcov.h>
#include <Map/mapfclass.h>
#include <Map/mapfeature.h>
#include <Map/mapprim.h>

#include <OverlayUnidraw/paramlist.h>
#include <OverlayUnidraw/ovclasses.h>
#include <OverlayUnidraw/ovfile.h>

#include <UniIdraw/idarrows.h>

#include <Unidraw/Graphic/ellipses.h>
#include <Unidraw/Graphic/lines.h>
#include <Unidraw/Graphic/polygons.h>
#include <Unidraw/Commands/command.h>

#include <Unidraw/catalog.h>
#include <Unidraw/editor.h>
#include <Unidraw/iterator.h>
#include <Unidraw/ulist.h>
#include <Unidraw/unidraw.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <TopoFace/fgeomobjs.h>

#include <OS/math.h>

#include <iostream.h>
#include <stdio.h>

extern "C" {
#include <projects.h>
}


/*****************************************************************************/

MapFeatureScript::MapFeatureScript (MapFeatureComp* subj) : OverlayScript(subj) {}

ClassId MapFeatureScript::GetClassId () { return MAPFEATURE_SCRIPT; }

boolean MapFeatureScript::IsA (ClassId id) {
    return MAPFEATURE_SCRIPT == id || OverlayScript::IsA(id);
}

boolean MapFeatureScript::Definition (ostream& out) {
    MapFeatureComp* comp = (MapFeatureComp*)GetSubject();
    Command *cmd = GetCommand();
    MapFeatureView* view;
    MapFeature* feature;
    ExportScriptType exporttype;
    if (cmd != nil) {
        if (cmd->IsA(MAPEXPORT_CMD)) {
            exporttype = ((MapExportCmd*)cmd)->type();
            feature = comp->feature();
	    Viewer* viewer = cmd->GetEditor()->GetViewer();
	    view = (MapFeatureView*)comp->FindView(viewer);
        } else
	    return 0;
    }

    Graphic* gr = view->GetGraphic();
    const Coord* x, *y;
    Coord x0, y0;
    int n, i, r1, r2;
    switch(feature->type()) {
    case MapFeature::PointType:
	out << "    ellipse(";
	((SF_Ellipse*)gr)->GetOriginal(x0, y0, r1, r2);
        out << x0 << "," << y0 << "," << r1 << "," << r2;
	break;
    case MapFeature::LineType:
	out << "    multiline(";
	n = ((SF_MultiLine*)gr)->GetOriginal(x, y);
	for (i=0; i<n; i++) {
	    out << "(" << x[i] << "," << y[i] << ")";
	    if (i+1 < n ) out << ",";
        }
	break;
    case MapFeature::AreaType:
	out << "    polygon(";
	n = ((SF_Polygon*)gr)->GetOriginal(x, y);
	for (i=0; i<n; i++) {
	    out << "(" << x[i] << "," << y[i] << ")";
	    if (i+1 < n ) out << ",";
        }
	break;
    case MapFeature::TextType:
	out << "    text(";
	break;
    case MapFeature::JunctionType:
	out << "    ellipse(";
	((SF_Ellipse*)gr)->GetOriginal(x0, y0, r1, r2);
        out << x0 << "," << y0 << "," << r1 << "," << r2;
	break;
    case MapFeature::RasterType:
    case MapFeature::DemType:
    default:
	cerr << "Unsupported (as of yet) map feature\n";
	return 0;
    }
    FillBg(out);
    MinGS(out);
    Annotation(out);
    out << " :featuretype " << feature->type();
    feature->definition(out);
    out << ")" << endl;
    return out.good();
}    

int MapFeatureScript::ReadFeature (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    MapFeature* feature = *(MapFeature**)addr1;

    boolean bad = true;
    char ch;

    switch(feature->type()) {

    case MapFeature::PointType:
    case MapFeature::LineType:
    case MapFeature::AreaType:
    case MapFeature::JunctionType:
        float *x, *y;
	int n;
        ch = in.peek();
        if (ch != ')' && ch  != ':') 
	    bad = ParamList::parse_fltpts(in, x, y, n);
        else {
	    x = y = nil;
	    n =0;
	    bad = false;
        }

        if (in.good() && !bad) { 
	  
            MapPrimitive* prim;
	    if (feature->type() == MapFeature::LineType ||
		feature->type() == MapFeature::PointType) 
	        prim = new MapEdge(feature);
	    else if (feature->type() == MapFeature::JunctionType)
	      prim = new MapNode(feature);
	    else
	        prim = new MapFace(feature);

   	    prim->insert_points(n, x, y);
	    feature->primitive(prim);

        } 
        delete x;
	delete y;
	break;

    case MapFeature::TextType:
	break;

    case MapFeature::RasterType:
    case MapFeature::DemType:
	cerr << "Unsupported map feature type\n";
	break;
    default:
	cerr << "Unknown map feature type\n";
	break;
    }

    if (bad) {
        cerr << "abnormal exit from MapFeatureScript::ReadPoints\n";
        return -1;
    }
    return 0;
}

/*****************************************************************************/

MapFClassScript::MapFClassScript (MapFClassComp* subj) : OverlaysScript(subj) {}

ClassId MapFClassScript::GetClassId () { return MAPFCLASS_SCRIPT; }

boolean MapFClassScript::IsA (ClassId id) {
    return MAPFCLASS_SCRIPT == id || OverlaysScript::IsA(id);
}

boolean MapFClassScript::Definition (ostream& out) {
    Iterator i;
    boolean status = true;
    Command *cmd = GetCommand();
    ExportScriptType exporttype;
    MapFClassComp* comp = (MapFClassComp*)GetSubject();

    boolean selected = true;
    if (cmd != nil && cmd->IsA(MAPEXPORT_CMD)) {
        exporttype = ((MapExportCmd*)cmd)->type();
	Viewer* viewer = cmd->GetEditor()->GetViewer();
	GraphicView* view = comp->FindView(viewer);
	selected = ((MapExportCmd*)cmd)->Selected(view);
    }  else {
        MapFeatureClass* fclass = comp->GetFeatureClass();
        out << "mapfcl( :path \"" << comp->GetPathName() << "\"";
	MinGS(out);
	Annotation(out);
	Attributes(out);
	out << ")";
	return status;
    }

    if (selected) {

        Indent(out);
        if (exporttype == MapIdrawType) {
	    MapFeatureClass* fclass = comp->GetFeatureClass();
            const char* name = fclass->name();

            out << "mapfcl(";
	    out << "\"" << name << "\"";

	} else {
            out << "picture(\n";
	}
    }
    Viewer* viewer = cmd->GetEditor()->GetViewer();
    if (cmd != nil && cmd->IsA(MAPEXPORT_CMD)) {
	for (First(i); status && !Done(i); ) {
	    OverlayScript* sv = (OverlayScript*)GetView(i);
	    MapFeatureView* view = (MapFeatureView*)comp->FindView(viewer);
	    boolean subselected = ((MapExportCmd*)cmd)->Selected(view);
	    if (subselected) {
		Indent(out);
	        status = sv->Definition(out);
	    }
	    Next(i);
	    if (subselected && !Done(i)) out << ",\n";
	}
    }
    if (selected) {
        FullGS(out);
        Annotation(out);
        out << ")";
    }
    return status;
}    

OverlayComp* MapFClassScript::read_obj(const char* name, istream& in, OverlaysComp* comps) {
  OverlayComp* child = nil;
  if (strcmp(name, "point") == 0 )
    child = new MapFeatureComp(in, MapFeature::PointType, comps);
  else if (strcmp(name, "line") == 0 )          
    child = new MapFeatureComp(in, MapFeature::LineType, comps);
  else if (strcmp(name, "area") == 0 )
    child = new MapFeatureComp(in, MapFeature::AreaType, comps);
  else if (strcmp(name, "text") == 0 )
    child = new MapFeatureComp(in, MapFeature::TextType, comps);
  else if (strcmp(name, "junction") == 0 )
    child = new MapFeatureComp(in, MapFeature::JunctionType, comps);
  else if (strcmp(name, "route") == 0 )
    child = new MapRouteComp(in, comps);
  else 
    child = OverlaysScript::read_obj(name, in, comps);
  return child;
}

int MapFClassScript::ReadPath (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    char pathname[BUFSIZ];

    MapFClassComp* comp = (MapFClassComp*)addr1;
    ParamList::parse_pathname(in, pathname, BUFSIZ, comp->GetBaseDir());

    if (!in.good()) {
	cerr << "Error in reading pathname for map feature class\n";
	return -1;
    } else {
        comp->SetPathName(pathname);
        return 0;
    }
}

int MapFClassScript::ReadFile (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
  return ReadPath(in, addr1, addr2, addr3, addr4);
}

/*****************************************************************************/

MapCovScript::MapCovScript (MapCovComp* subj) : OverlaysScript(subj) {}

ClassId MapCovScript::GetClassId () { return MAPCOV_SCRIPT; }

boolean MapCovScript::IsA (ClassId id) {
    return MAPCOV_SCRIPT == id || OverlaysScript::IsA(id);
}

boolean MapCovScript::Definition (ostream& out) {
    Iterator i;
    boolean status = true;
    Command *cmd = GetCommand();
    ExportScriptType exporttype;
    MapCovComp* comp = (MapCovComp*)GetSubject();
    boolean selected = true;
    if (cmd != nil && cmd->IsA(MAPEXPORT_CMD)) {
        exporttype = ((MapExportCmd*)cmd)->type();
	Viewer* viewer = cmd->GetEditor()->GetViewer();
	GraphicView* view = comp->FindView(viewer);
	selected = ((MapExportCmd*)cmd)->Selected(view);
    }
    else
        exporttype = MapIdrawType;
    if (selected && exporttype == MapIdrawType) {
        MapCoverage* cov = comp->GetCoverage();
        const char* name = cov->name();

        out << "mapcov(";
	out << "\"" << name << "\"\n";
    }

    for (First(i); status && !Done(i); ) {
	OverlayScript* sv = (OverlayScript*)GetView(i);
        if (selected && exporttype != OvIdrawType)
  	    Indent(out);
        status = sv->Definition(out);
	Next(i);
	if (!Done(i) && selected) out << ",\n";
    }
    if (selected && exporttype != OvIdrawType) {
        out << "\n";
        Indent(out);
        FullGS(out);
        Annotation(out);
        out << ")";
    }
    return status;
}    

/*****************************************************************************/

MapLibScript::MapLibScript (MapLibComp* subj) : OverlaysScript(subj) {}

ClassId MapLibScript::GetClassId () { return MAPLIB_SCRIPT; }

boolean MapLibScript::IsA (ClassId id) {
    return MAPLIB_SCRIPT == id || OverlaysScript::IsA(id);
}

boolean MapLibScript::Definition (ostream& out) {
    Iterator i;
    boolean status = true;
    Command *cmd = GetCommand();
    ExportScriptType exporttype;
    MapLibComp* comp = (MapLibComp*)GetSubject();
    boolean selected = true;
    if (cmd != nil && cmd->IsA(MAPEXPORT_CMD)) {
        exporttype = ((MapExportCmd*)cmd)->type();
	Viewer* viewer = cmd->GetEditor()->GetViewer();
	GraphicView* view = comp->FindView(viewer);
	selected = ((MapExportCmd*)cmd)->Selected(view);
    }
    else
        exporttype = MapIdrawType;

    if (selected && exporttype == MapIdrawType) {
        MapLibrary* lib = comp->GetLibrary();
        const char* name = lib->name();

        out << "maplib(";
	out << "\"" << name << "\"\n";
    }

    for (First(i); status && !Done(i); ) {
	OverlayScript* sv = (OverlayScript*)GetView(i);
        if (selected && exporttype != OvIdrawType)
 	    Indent(out);
        status = sv->Definition(out);
	Next(i);
	if (selected && !Done(i)) out << ",\n";
    }
    if (selected && exporttype != OvIdrawType) {
        out << "\n";
        Indent(out);
        FullGS(out);
        Annotation(out);
        out << ")";
    }
    return status;
}    

/*****************************************************************************/

MapDbScript::MapDbScript (MapDbComp* subj) : OverlaysScript(subj) {}

ClassId MapDbScript::GetClassId () { return MAPDB_SCRIPT; }

boolean MapDbScript::IsA (ClassId id) {
    return MAPDB_SCRIPT == id || OverlaysScript::IsA(id);
}

boolean MapDbScript::Definition (ostream& out) {
    Iterator i;
    boolean status = true;
    Command *cmd = GetCommand();
    MapDbComp* comp = (MapDbComp*)GetSubject();
    boolean selected = true;
    ExportScriptType exporttype;
    if (cmd != nil && cmd->IsA(MAPEXPORT_CMD)) {
        exporttype = ((MapExportCmd*)cmd)->type();
	Viewer* viewer = cmd->GetEditor()->GetViewer();
	GraphicView* view = comp->FindView(viewer);
	selected = ((MapExportCmd*)cmd)->Selected(view);
    }
    else
        exporttype = MapIdrawType;

    if (selected && exporttype == MapIdrawType) {
        out << "mapdb(";
        MapDatabase* db = comp->GetDatabase();
        const char* path = db->path();
	out << "\"" << path << "\"\n";
    }
    for (First(i); status && !Done(i); ) {
	OverlayScript* sv = (OverlayScript*)GetView(i);
	if (selected && exporttype != OvIdrawType)
            Indent(out);
        status = sv->Definition(out);
	Next(i);
	if (selected && !Done(i)) out << ",\n";
    }
    if (selected && exporttype != OvIdrawType) {
        out << "\n";
        Indent(out);
        FullGS(out);
        Annotation(out);

        out << ")";
    }
    return status;
}    

/*****************************************************************************/

MapIdrawScript::MapIdrawScript (MapIdrawComp* subj) : FrameIdrawScript(subj) {}

ClassId MapIdrawScript::GetClassId () { return MAP_IDRAW_SCRIPT; }

boolean MapIdrawScript::IsA (ClassId id) { 
    return MAP_IDRAW_SCRIPT == id || FrameIdrawScript::IsA(id);
}

boolean MapIdrawScript::Emit (ostream& out) {
    Command *cmd = GetCommand();
    ExportScriptType exporttype;
    if (cmd != nil && cmd->IsA(MAPEXPORT_CMD))
        exporttype = ((MapExportCmd*)cmd)->type();
    else
	exporttype = MapIdrawType;

    if (exporttype == MapIdrawType)
        out << "mapserv(";
    if (exporttype == OvIdrawType)
        out << "drawtool(";

    Iterator i;
    boolean status = true;
    First(i);
    if (!Done(i)) out << "\n";
    for (; status && !Done(i); ) {
	ExternView* ev = GetView(i);
        if (exporttype != OvIdrawType)
	    Indent(out);
        status = ev->Definition(out);
	Next(i);
	if (!Done(i)) out << ",\n";
    }

    out << "\n";
    FullGS(out);
    Annotation(out);
    out << ")\n";
    return status;
}

/*****************************************************************************/

boolean MapFeatureRawScript::_native = true;

MapFeatureRawScript::MapFeatureRawScript (MapFeatureComp* subj) : MapFeatureScript(subj) {}

ClassId MapFeatureRawScript::GetClassId () { return MAPFEATURE_RAW; }

boolean MapFeatureRawScript::IsA (ClassId id) {
    return MAPFEATURE_RAW == id || MapFeatureScript::IsA(id);
}

boolean MapFeatureRawScript::Definition (ostream& out) {
    MapFeatureComp* comp = (MapFeatureComp*)GetSubject();
    MapFeature* feature = comp->feature();
    int i;
    switch(feature->type()) {
    case MapFeature::PointType:
	{
	    MapPoint* point = feature->point_primitive();
	    if (!point) break;
	    if (MapFeatureRawScript::native()) 
	      out.form("point(%.4f,%.4f", point->xorigin(), point->yorigin());
	    else
	        out << "ellipse(" << Math::round(point->xorigin()) << ","
		    << Math::round(point->yorigin()) << "10,10";
	}
	break;
    case MapFeature::LineType:
	{
	    MapEdge* edge = feature->edge_primitive();
	    if (!edge) break;
	    int npts = edge->npts();
	    const float* fx = edge->xpoints();
	    const float* fy = edge->ypoints();
	    if (MapFeatureRawScript::native()) 
	        out << "line(";
	    else
	        out << "multiline(";
	    for (i=0; i<npts; i++) {
                if (MapFeatureRawScript::native()) 
		  out.form("(%.4f,%.4f)", fx[i], fy[i]);
                else
		    out << "(" << Math::round(fx[i]) << "," << Math::round(fy[i]) << ")";
		if (i+1 < npts ) out << ",";
	    }
	}
	break;
    case MapFeature::AreaType:
	{
	    MapFace* face = feature->face_primitive();
	    if (!face) break;
	    int npts = face->npts();
	    const float* fx = face->xpoints();
	    const float* fy = face->ypoints();
	    if (MapFeatureRawScript::native()) 
	        out << "area(";
            else 
	        out << "polygon(";
	    for (i=0; i<npts; i++) {
                if (MapFeatureRawScript::native())
		  out.form("(%.4f,%.4f)", fx[i], fy[i]);
                else
		    out << "(" << Math::round(fx[i]) << "," << Math::round(fy[i]) << ")";
		if (i+1 < npts ) out << ",";
	    }
	}
	break;
    case MapFeature::TextType:
	{
	    MapText* text = feature->text_primitive();
	    if (!text) break;
	    out.form("text(%.4f,%.4f", text->xorigin(), text->yorigin());
	    out << ", \"" << text->text() << "\"";
	}
	break;
    case MapFeature::JunctionType:
	{
	    MapNode* node = feature->node_primitive();
	    if (!node) break;
	    if (MapFeatureRawScript::native()) 
	      out.form("junction(%.4f,%.4f", node->xorigin(), node->yorigin());
	    else
	        out << "ellipse(" << Math::round(node->xorigin()) << ","
		    << Math::round(node->yorigin()) << "10,10";
	}
	break;
    case MapFeature::RasterType:
    case MapFeature::DemType:
    default:
	cerr << "Unsupported (as of yet) map feature\n";
	break;
    }

    out << " :featuretype " << feature->type();
    //    feature->definition(out);
    if (feature->attrlist() && feature->attrlist() != comp->GetAttributeList())
      out << *feature->attrlist();
    out << *comp->GetAttributeList();
    out << ")" << endl;
    return out.good();
}    

/*****************************************************************************/

MapFClassRawScript::MapFClassRawScript (MapFClassComp* subj) : MapFClassScript(subj) {}

ClassId MapFClassRawScript::GetClassId () { return MAPFCLASS_RAW; }

boolean MapFClassRawScript::IsA (ClassId id) {
    return MAPFCLASS_RAW == id || MapFClassScript::IsA(id);
}

static int utmzone(float lon) {
  if (lon > -180.0 && lon < -174.0)
    return 1;
  else if (lon > -174.0 && lon < -168.0)
    return 2;
  else if (lon > -168.0 && lon < -162.0)
    return 3;
  else if (lon > -162.0 && lon < -156.0)
    return 4;
  else if (lon > -156.0 && lon < -150.0)
    return 5;
  else if (lon > -150.0 && lon < -144.0)
    return 6;
  else if (lon > -144.0 && lon < -138.0)
    return 7;
  else if (lon > -138.0 && lon < -132.0)
    return 8;
  else if (lon > -132.0 && lon < -126.0)
    return 9;
  else if (lon > -126.0 && lon < -120.0)
    return 10;
  else if (lon > -120.0 && lon < -114.0)
    return 11;
  else if (lon > -114.0 && lon < -108.0)
    return 12;
  else if (lon > -108.0 && lon < -102.0)
    return 13;
  else if (lon > -102.0 && lon < -96.0)
    return 14;
  else if (lon > -96.0 && lon < -90.0)
    return 15;
  else if (lon > -90.0 && lon < -84.0)
    return 16;
  else if (lon > -84.0 && lon < -78.0)
    return 17;
  else if (lon > -78.0 && lon < -72.0)
    return 18;
  else if (lon > -72.0 && lon < -66.0)
    return 19;
  else if (lon > -66.0 && lon < -60.0)
    return 20;
  else if (lon > -60.0 && lon < -54.0)
    return 21;
  else if (lon > -54.0 && lon < -48.0)
    return 22;
  else if (lon > -48.0 && lon < -42.0)
    return 23;
  else if (lon > -42.0 && lon < -36.0)
    return 24;
  else if (lon > -36.0 && lon < -30.0)
    return 25;
  else if (lon > -30.0 && lon < -24.0)
    return 26;
  else if (lon > -24.0 && lon < -18.0)
    return 27;
  else if (lon > -18.0 && lon < -12.0)
    return 28;
  else if (lon > -12.0 && lon < -6.0)
    return 29;
  else if (lon > -6.0 && lon < 0.0)
    return 30;
  else if (lon > 0.0 && lon < 6.0)
    return 31;
  else if (lon > 6.0 && lon < 12.0)
    return 32;
  else if (lon > 12.0 && lon < 18.0)
    return 33;
  else if (lon > 18.0 && lon < 24.0)
    return 34;
  else if (lon > 24.0 && lon < 30.0)
    return 35;
  else if (lon > 30.0 && lon < 36.0)
    return 36;
  else if (lon > 36.0 && lon < 42.0)
    return 37;
  else if (lon > 42.0 && lon < 48.0)
    return 38;
  else if (lon > 48.0 && lon < 54.0)
    return 39;
  else if (lon > 54.0 && lon < 60.0)
    return 40;
  else if (lon > 60.0 && lon < 66.0)
    return 41;
  else if (lon > 66.0 && lon < 72.0)
    return 42;
  else if (lon > 72.0 && lon < 78.0)
    return 43;
  else if (lon > 78.0 && lon < 84.0)
    return 44;
  else if (lon > 84.0 && lon < 90.0)
    return 45;
  else if (lon > 90.0 && lon < 96.0)
    return 46;
  else if (lon > 96.0 && lon < 102.0)
    return 47;
  else if (lon > 102.0 && lon < 108.0)
    return 48;
  else if (lon > 108.0 && lon < 114.0)
    return 49;
  else if (lon > 114.0 && lon < 120.0)
    return 50;
  else if (lon > 120.0 && lon < 126.0)
    return 51;
  else if (lon > 126.0 && lon < 132.0)
    return 52;
  else if (lon > 132.0 && lon < 138.0)
    return 53;
  else if (lon > 138.0 && lon < 144.0)
    return 54;
  else if (lon > 144.0 && lon < 150.0)
    return 55;
  else if (lon > 150.0 && lon < 156.0)
    return 56;
  else if (lon > 156.0 && lon < 162.0)
    return 57;
  else if (lon > 162.0 && lon < 168.0)
    return 58;
  else if (lon > 168.0 && lon < 174.0)
    return 59;
  else if (lon > 174.0 && lon < 180.0)
    return 60;
  else
    return 0;
}

boolean MapFClassRawScript::Definition (ostream& out) {
    Iterator i;
    boolean status = true;
    ExportScriptType exporttype;
    MapFClassComp* comp = (MapFClassComp*)GetSubject();
    boolean selected = true;
    exporttype = MapIdrawType;

    MapFeatureClass* fclass = comp->GetFeatureClass();
    const char* name = fclass->name();

    for (First(i); status && !Done(i); ) {
      OverlayScript* sv = (OverlayScript*)GetView(i);
      if (selected && exporttype != OvIdrawType)	    
	Indent(out);
      status = sv->Definition(out);
      Next(i);
      if (selected && !Done(i)) out << ",\n";
    }

    int utm_zone = 0;
    char* parms[4];
    char lon0[80];
      parms[0] = "proj=utm";
      parms[1] = "ellps=WGS84";
      parms[3] = "no_defs";
    double x1, y1, x2, y2;
    if (!comp->GetAttributeList()->GetAttr("zone")) {
      fclass->extent(x1, y1, x2, y2, utm_zone);
      //    out << "\n :ext_x1 " << x1 << " :ext_y1 " << y1;
      //    out << " :ext_x2 " << x2 << " :ext_y2 " << y2;
      // Do it this way instead so floats come out dotted
      //    out.form("\n :ext_x1 %.4f :ext_y1 %.4f :ext_x2 %.4f :ext_y2 %.4f",
      //	     x1, y1, x2, y2);
      if (utm_zone == 0) {
	sprintf(lon0, "lon_0=%f", x1);  // let proj pick the utm zone
	parms[2] = lon0;
    
	//    char ln0[80];
	//    sprintf(ln0, "%f", x1);
	utm_zone = utmzone(x1+0.1);
      }
      //    out << " :lon0 " << ln0;
      out << "\n :proj " << "utm";
      out << " :ellps " << "WGS84";
      out << " :zone " << utm_zone;

      if (utm_zone == 0) {
	PJ* proj = pj_init(4, parms);
	UV data;
	data.u = x1 * DEG_TO_RAD;
	data.v = y1 * DEG_TO_RAD;
	data = pj_fwd(data, proj);
	float px1 = data.u;
	float py1 = data.v;
	data.u = x2 * DEG_TO_RAD;
	data.v = y2 * DEG_TO_RAD;
	data = pj_fwd(data, proj);
	float px2 = data.u;
	float py2 = data.v;
      }
      out.form("\n :sw_e %.6f :sw_n %.6f :ne_e %.6f :ne_n %.6f",
#if 0
	     px1, py1, px2, py2);  // project extent
#else
	     x1, y1, x2, y2); // already projected extent
#endif
    }
    out << "\n :mapfc \"" << fclass->name() << "\"";
    out << " :mapfc_desc \"" << fclass->description() << "\"";
    MapCoverage* cov = fclass->coverage();
    if (cov) {
        out << " :mapcov \"" << cov->name() << "\"";
        out << " :mapcov_desc \"" << cov->description() << "\"";
        MapLibrary* lib = cov->library();
	if (lib) {
            out << " :maplib \"" << lib->name() << "\"";
            out << " :maplib_desc \"" << lib->description() << "\"";
            MapDatabase* db = lib->database();
	    if (db) {
                out << " :mapdb \"" << db->name() << "\"";
                out << " :mapdb_desc \"" << db->description() << "\"";
  	    }
	}
    }
    return status;
}    

void MapFClassRawScript::Update() {
    DeleteViews();
    OverlayComp* comps;

    MapFClassComp* comp = (MapFClassComp*)GetSubject();
    MapFeatureClass* fclass = comp->GetFeatureClass();

    comps = GetOverlayComp();
    Iterator i;

    for (comps->First(i); !comps->Done(i); comps->Next(i)) {
        OverlayComp* comp = (OverlayComp*) comps->GetComp(i);
        OverlayScript* sv = CreateRawMapScript(comp);

        if (sv != nil) {
            _views->Append(new UList(sv));
	    SetParent(sv, this);
	}
    }

}

OverlayScript* MapFClassRawScript::CreateRawMapScript (OverlayComp* comp) {
    OverlayScript* sv = (OverlayScript*) comp->Create(RAWMAP_VIEW);

    if (sv != nil) {
        comp->Attach(sv);
        sv->Update();
    }
    return sv;
}
